from asyncio import tasks
import requests
import re
import time
import tehran_stocks.config as db
from tehran_stocks.models import Stocks


def get_stock_ids():
    url = "http://tsetmc.com/tsev2/data/MarketWatchPlus.aspx"
    r = requests.get(url)
    ids = set(re.findall(r"\d{15,20}", r.text))
    return list(ids)


def get_stock_groups():
    """
    group numbers from tccim, to avoid searching useless group numbers.
    its a helper for other parts of package to collect stock lists.
    """
    r = requests.get("http://www.tsetmc.com/Loader.aspx?ParTree=111C1213")
    return re.findall(r"\d{2}", r.text)


def create_or_update_stock_from_dict(stock_id, stock):
    if exist := Stocks.query.filter_by(code=stock_id).first():
        print(f"stock with code {stock_id} exist")
        exist.shareCount = stock["shareCount"]
        exist.baseVol = stock["baseVol"]
        exist.sectorPe = stock["sectorPe"]
        exist.estimatedEps = stock["estimatedEps"]
    else:
        print(f"creating stock with code {stock_id}")
        db.session.add(Stocks(**stock))


def get_stock_detail(stock_id: str) -> Stocks:
    """
    Dowload stocks detail and save them to the database.
    better not use it alone.
    Its not useful after first setup. ;)


    params
    ----------------
    stockk_id :
        an str of an interger or an integer if not started by 0 whicj represent  the Id
        in tsetmc website i.e. arg i={stock_id} inside  the url
    group_id:
        int: number that represent group of stock

    """

    url = f"http://www.tsetmc.com/Loader.aspx?ParTree=151311&i={stock_id}"
    r = requests.get(url)
    stock = {
        "code": stock_id,
        "instId": re.findall(r"InstrumentID='([\w\d]*)|$',", r.text)[0],
    }

    stock["insCode"] = (
        stock_id if re.findall(r"InsCode='(\d*)',", r.text)[0] == stock_id else 0
    )
    stock["baseVol"] = float(re.findall(r"BaseVol=([\.\d]*),", r.text)[0])
    try:
        stock["name"] = re.findall(r"LVal18AFC='([\D]*)',", r.text)[0]
    except:
        return
    try:
        stock["group_name"] = re.findall(r"LSecVal='([\D]*)',", r.text)[0]
    except:
        return
    try:
        stock["title"] = re.findall(r"Title='([\D]*)',", r.text)[0]
    except:
        return
    try:
        stock["sectorPe"] = float(re.findall(r"SectorPE='([\.\d]*)',", r.text)[0])
    except:
        stock["sectorPe"] = None
    try:
        stock["shareCount"] = float(re.findall(r"ZTitad=([\.\d]*),", r.text)[0])
    except:
        stock["shareCount"] = None

    try:
        stock["estimatedEps"] = float(
            re.findall(r"EstimatedEPS='([\.\d]*)',", r.text)[0]
        )
    except:
        stock["estimatedEps"] = None
    stock["group_code"] = re.findall(r"CSecVal='([\w\d]*)|$',", r.text)[0]
    if stock["name"] == "',DEven='',LSecVal='',CgrValCot='',Flow='',InstrumentID='":
        return False

    create_or_update_stock_from_dict(stock_id, stock)

    try:
        db.session.commit()
    except:
        print(f"stock with code {stock_id} exist")
        db.session.rollback()
    return stock


def fill_stock_table():
    """
    Download Stock Table,
    1- gets groups
    2- gets stock in groups
    3- download stock detail
    4- save them to database
    5- guides you to use the package
    """
    URL = "https://ts-api.ir/api/stocks"
    try:
        print("try Downloading stock table from the package api... ")
        print("visit: https://ts-api.ir/")
        r = requests.get(URL)
        if r.status_code != 200:
            raise ConnectionError("connection error")
        data = r.json()
        stocks = data["stocks"]
        for stock in stocks:
            del stock["id"]
            create_or_update_stock_from_dict(stock["code"], stock)
        db.session.commit()
        return

    except ConnectionError:
        print("fall back to manual mode")
        pass
    except Exception as e:
        print(e)
        pass
    print("Downloading group ids...")
    stocks = get_stock_ids()
    for i, stock in enumerate(stocks):
        get_stock_detail(stock)
        print(
            f"downloading stocks details, changes: {(i+1)/len(stocks)*100:.1f}% completed",
            end="\r",
        )

    print("Add all groups, you can download stock price by following codes")
    print("from tehran_stocks import downloader")
    print(" downloader.download_all() # for downloading all data")
    print("downloader.download_group(group_id) # to download specefic group data")
    print("downloader.download_stock(stock) to downloand stock specefic")



import os

import tehran_stocks.config as db
from tehran_stocks.download import (
    get_all_price,
    get_stock_detail,
    get_stock_groups,
    get_stock_ids,
    update_group,
    update_stock_price,
)
from tehran_stocks.models import StockPrice, Stocks, get_asset

from .initializer import init_db, fill_db


def db_is_empty():
    try:
        db.session.execute("select * from stocks limit 1;")
        return False
    except:
        return True


if db_is_empty():
    print("No database founded.")
    init_db()
    fill_db()



import os

from tehran_stocks import db, models, update_group, get_all_price
from tehran_stocks.download import fill_stock_table


def init_db():
    print("creating database")
    path = os.path.join(db.HOME_PATH, db.TSE_FOLDER)

    if "tse" not in os.listdir(db.HOME_PATH):
        print("making package folder...")
        print("Includes: config.yml and stocks.db  if you are using sqlite.")
        print("you can change config.yml to your needs.")
        try:
            os.mkdir(path)
        except FileExistsError:
            print("folder already exists")
    models.create()
    print(f"DataBase created in: {path}")


def fill_db():
    print("downloading stock name and details from TSETMC")
    print("may take few minutes")

    fill_stock_table()
    print("Stock table is available now, example:")
    print("from tehran_stocks import Stocks")
    print('stock =Stocks.query.filter_by(name="کگل").first()')

    a = input("Do you want to download all price? [y,(n)]")
    if a == "y":
        print("Downloading price:")
        get_all_price()
    else:
        print("if  you want download all prices use tehran_stocks.get_all_price() ")
        print("if you want download price history of a specfic stock use: ")
        print("stock.update()")
        print("or use tehran_stocks.update_group(id) ")
        print("For more info go to:")
        print("https://github.com/ghodsizadeh/tehran-stocks")



from tehran_stocks.config import *
from sqlalchemy.orm import relationship
import pandas as pd
import requests
import jalali_pandas


class Stocks(Base):
    __tablename__ = "stocks"

    id = Column(Integer, primary_key=True)
    name = Column(String)
    title = Column(String)
    group_name = Column(String)
    group_code = Column(Integer)
    instId = Column(String)
    insCode = Column(String)
    code = Column(String, unique=True)
    sectorPe = Column(Float)
    shareCount = Column(Float)
    estimatedEps = Column(Float)
    baseVol = Column(Float)
    prices = relationship("StockPrice", backref="stock")
    _cached = False
    _dfcounter = 0

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def df(self) -> pd.DataFrame:
        """dataframe of stock price with date and OHLC"""
        self._dfcounter += 1
        if self._cached:
            return self._df
        query = f"select * from stock_price where code = '{self.code}'"
        df = pd.read_sql(query, engine)
        if df.empty:
            self._cached = True
            self._df = df
            return self._df
        df["date"] = pd.to_datetime(df["dtyyyymmdd"], format="%Y%m%d")
        df["jdate"] = df.date.jalali.to_jalali()
        df = df.sort_values("date")
        df.reset_index(drop=True, inplace=True)
        df.set_index("date", inplace=True)
        self._cached = True
        self._df = df

        return self._df

    def get_dividend(self) -> pd.DataFrame:
        """get changes in price for dividend and changes in share
        postive value is dividend and negative value is changes in share

        Returns:
            pd.DataFrame: _description_
        """

        url = f"http://www.tsetmc.com/Loader.aspx?Partree=15131G&i={self.code}"
        r = requests.get(url)
        changes = pd.read_html(r.text)[0]
        changes.columns = ["date", "after", "before"]
        changes["dividend"] = changes.before - changes.after
        changes["date"] = changes.date.jalali.parse_jalali("%Y/%m/%d")
        changes["gdate"] = changes.date.jalali.to_gregorian()
        return changes

    def get_shares_history(self) -> pd.DataFrame:
        """_summary_get changes in shares

        Returns:
            pd.DataFrame: return day of shares changes and shares count [date, new_shares, old_shares, gdate]
        """
        url = f"http://www.tsetmc.com/Loader.aspx?Partree=15131H&i={self.code}"
        r = requests.get(url)
        df = pd.read_html(r.text)[0]
        df.columns = ["date", "new_shares", "old_shares"]
        df["date"] = df.date.jalali.parse_jalali("%Y/%m/%d")
        df["gdate"] = df.date.jalali.to_gregorian()
        return df

    @property
    def mpl(self):
        self._mpl = self.df.rename(
            columns={
                "close": "Close",
                "open": "Open",
                "high": "High",
                "low": "Low",
                "vol": "Volume",
            }
        )
        return self._mpl

    def update(self):
        from tehran_stocks.download import update_stock_price

        try:
            return update_stock_price(self.code)
        except:
            return False

    def summary(self):
        """summart of stock"""
        df = self.df
        sdate = df.index.min().strftime("%Y/%m/%d")
        edate = df.index.max().strftime("%Y/%m/%d")

        print(f"Start date: {sdate}")
        print(f"End date: {edate}")
        print(f"Total days: {len(df)}")

    def get_instant_detail(self) -> dict:
        """get instant detail of stock
        last_price, last_close, last_open, last_high, last_low, last_vol, trade_count, trade_value,market_cap
        instantly from the website

        Returns:
            dict: { last_price, last_close, last_open, last_high, last_low, last_vol, trade_count, trade_value,market_cap}
        """
        url = (
            f"http://www.tsetmc.com/tsev2/data/instinfodata.aspx?i={self.code}&c=27%20"
        )
        headers = {
            "Connection": "keep-alive",
            "Accept": "text/plain, */*; q=0.01",
            "DNT": "1",
            "X-Requested-With": "XMLHttpRequest",
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.74 Safari/537.36",
        }
        # 12:29:40,A ,4300,4364,4577,4361,4577,4250,195,702025,3070743824,0,20220326,122940;
        r = requests.get(url, headers=headers)
        main_response = r.text.split(";")[0]
        main_response = main_response.split(",")

        self.time = main_response[0]
        self.last_price = main_response[2]
        self.last_close = main_response[3]
        self.last_high = main_response[4]
        self.last_low = main_response[5]
        self.last_open = main_response[6]
        self.trade_count = main_response[7]
        self.trade_volume = main_response[8]
        self.trade_value = main_response[9]
        self.market_cap = main_response[10]
        self.date_string = main_response[12]
        self.time_string = main_response[13]
        del main_response[11]
        del main_response[1]

        keys = [
            "time",
            "last_price",
            "last_close",
            "last_high",
            "last_low",
            "last_open",
            "trade_count",
            "trade_volume",
            "trade_value",
            "market_cap",
            "date_string",
            "time_string",
        ]
        return dict(zip(keys, main_response))

    def __repr__(self):
        return f"{self.title}-{self.name}-{self.group_name}"

    def __str__(self):
        return self.name

    @staticmethod
    def get_group():
        return (
            session.query(Stocks.group_code, Stocks.group_name)
            .group_by(Stocks.group_code)
            .all()
        )


class StockPrice(Base):
    __tablename__ = "stock_price"

    id = Column(Integer, primary_key=True)
    code = Column(String, ForeignKey("stocks.code"), index=True)
    ticker = Column(String)
    date = Column("dtyyyymmdd", Integer, index=True)
    date_shamsi = Column(String)
    first = Column(Float)
    high = Column(Float)
    low = Column(Float)
    close = Column(Float)
    value = Column(BIGINT)
    vol = Column(BIGINT)
    openint = Column(Integer)
    per = Column(String)
    open = Column(Float)
    last = Column(Float)

    def __repr__(self):
        return f"{self.stock.name}, {self.date}, {self.close:.0f}"


def get_asset(name):
    name = name.replace("ی", "ي").replace("ک", "ك")
    return Stocks.query.filter_by(name=name).first()



from sqlalchemy import Column, Float, Integer, String, create_engine
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.orm import sessionmaker
from sqlalchemy import *
from pathlib import Path
import os
import sqlite3
import logging
import yaml


HOME_PATH = str(Path.home())
TSE_FOLDER = ".tse"
CONFIG_PATH = os.path.join(HOME_PATH, TSE_FOLDER) + "/" + "config.yml"

if not os.path.exists(CONFIG_PATH):
    # create config.yml from config.deafult.yml
    with open(os.path.join(os.path.dirname(__file__), "config.default.yml"), "r") as f:
        config = yaml.full_load(f)

    with open(CONFIG_PATH, "w") as f:
        yaml.dump(config, f)


with open(CONFIG_PATH, "r") as f:
    config = yaml.full_load(f)

defualt_db_path = os.path.join(f"{HOME_PATH}/{TSE_FOLDER}/stocks.db")

db_path = config.get("database").get("path")
if db_path is None:
    db_path = defualt_db_path

default_engine_URI = f"sqlite:///{db_path}"
engine = config.get("database").get("engine")
if engine == "sqlite":
    engine_URI = default_engine_URI
else:
    engine = config.get("database").get("engine")
    host = config.get("database").get("host")
    port = config.get("database").get("port")
    database = config.get("database").get("database")
    user = config.get("database").get("user")
    password = config.get("database").get("password")
    engine_URI = f"{engine}://{user}:{password}@{host}:{port}/{database}"


logging.info(engine_URI)
engine = create_engine(engine_URI)


Session = sessionmaker()
Session.configure(bind=engine)
session = Session()


class ClassProperty(object):
    def __init__(self, fget):
        self.fget = fget

    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)

    def __repr__(self):
        return None


class QueryMixin:
    @ClassProperty
    def query(cls):
        return session.query(cls)

    def display(self):
        data = self.__dict__
        try:
            del data["_sa_instance_state"]
        except:
            pass
        return data


Base = declarative_base(cls=QueryMixin)
